import bdb
import contextlib
import os
import time

import apprise


###############################################################################
# Send job notifications
###############################################################################


@contextlib.contextmanager
def notify(
    description: str,
    track_time: bool = True,
    notify_on_fail: bool = True):
    """Context manager for sending job notifications

    Arguments
        description - The name of the job being run
        track_time - Whether to report time elapsed
        notify_on_fail - Whether to send a notification on failure
    """
    # Start time
    if track_time:
        start_time = time.time()

    try:

        # Run user code
        yield

    except Exception as exception:

        # Ignore pdb exceptions
        if isinstance(exception, bdb.BdbQuit):
            return

        if notify_on_fail:

            # End time
            elapsed = time.time() - start_time if track_time else None

            # Report failure
            push_failure(description, track_time, elapsed, exception)

        raise exception

    # End time
    elapsed = time.time() - start_time if track_time else None

    # Report success
    push_success(description, track_time, elapsed)


###############################################################################
# Utilities
###############################################################################


def push(message: str):
    """Send a push notification to all of the configured services"""
    service = os.getenv('PYTORCH_NOTIFICATION_URL', default=None)
    if service is not None:
        if not hasattr(push, 'messenger'):
            push.messenger = apprise.Apprise()
            push.messenger.add(service)
        push.messenger.notify(message)


def push_failure(description, track_time, elapsed, exception):
    """Send notification of task failure"""
    if track_time:
        message = (
            f'Task "{description}" failed with '
            f'exception: {exception.__class__} in '
            f'{elapsed:.2f} seconds')
    else:
        message = (
            f'Task "{description}" failed with '
            f'exception: {exception.__class__}')
    push(message)


def push_success(description, track_time, elapsed):
    """Send notification of successful task completion"""
    if track_time:
        message = (
            f'Task "{description}" finished in '
            f'{elapsed:.2f} seconds')
    else:
        message = f'Task "{description}" finished'
    push(message)



import os
from pathlib import Path
from typing import List, Optional, Union

import torchutil


###############################################################################
# Constants
###############################################################################


# Available units of memory
UNITS = ['B', 'KB', 'MB', 'GB', 'TB']


###############################################################################
# Measure dataset sizes
###############################################################################


def measure(
    globs: Union[str, List[str]],
    roots: Optional[
        List[
            Union[
                Union[str, bytes, os.PathLike],
                List[Union[str, bytes, os.PathLike]]
            ]
        ]
    ] = None,
    recursive: bool = False,
    unit='B'
) -> Union[int, float]:
    """Measure data usage of files and directories

    Arguments
        globs
            Globs matching paths to measure
        roots
            Directories to apply glob searches; current directory by default
        recursive
            Apply globs to all subdirectories of root directories
        unit
            Unit of memory utilization (bytes to terabytes); default bytes
    """
    # Argument handling
    if isinstance(globs, str):
        globs = [globs]
    if not isinstance(roots, list):
        roots = [roots]
    roots = [Path() if root is None else Path(root) for root in roots]

    # Get paths to delete
    paths = []
    for root in roots:
        root = Path(root)
        for glob in globs:
            paths.extend(root.rglob(glob) if recursive else root.glob(glob))

    # Measure paths
    total = 0
    for path in torchutil.iterator(paths, 'Measuring paths'):
        total += Path(path).stat().st_size

    # Convert to desired unit of measurement
    return total if unit == 'B' else total / (1024 ** UNITS.index(unit))



from itertools import repeat
from pathlib import Path
from typing import List, Union, Optional
import os

def gather(
    *sources: List[Union[str, bytes, os.PathLike]],
    sinks: Optional[List[Union[str, bytes, os.PathLike]]] = None,
    source_extensions: Optional[str] = None,
    sink_extension: str = '.pt'):
    """
    Gathers lists of input and output directories and files into two lists
    of files, using the provided extension to glob directories.
    """
    if len(sources) == 0:
        raise ValueError('At least one source list must be provided')
    elif len(sources) == 1:
        assert sources[0] is not None, 'At least one source list must be provieded, but found None'
    else:
        assert source_extensions is None

    sources = list(sources)
    for i in range(0, len(sources)):
        if sources[i] is None:
            sources[i] = repeat(None)

    # Standardize extensions
    if source_extensions is not None:
        source_extensions = set([
            '.' + ext if '.' not in ext else ext
            for ext in source_extensions])
    else:
        source_extensions = []
    sink_extension = (
        '.' + sink_extension if '.' not in sink_extension
        else sink_extension)

    lengths = set()
    for source_list in sources:
        lengths.add(len(source_list))

    assert len(lengths) == 1, 'all source lists must have the same lengths'

    if sinks is not None:
        assert len(sinks) == len(sources[0]), 'sinks must have the same length as the source lists'

    if sinks is None:

        # Get sources as a list of files
        source_files = [[] for _ in sources]
        for source_tuple in zip(*sources):
            source_paths = [Path(source) for source in source_tuple]
            source_is_dir = list(set([source.is_dir() for source in source_paths]))
            if len(source_is_dir) > 1 and True in source_is_dir:
                raise ValueError('cannot handle directories when more than one input list is given')
            elif len(source_is_dir) == 1 and source_is_dir[0]:
                for extension in source_extensions:
                    source_files[0] += list(source_paths[0].rglob(f'*{extension}'))
            else:
                for i, source in enumerate(source_paths):
                    source_files[i].append(source)

        # Sink files are source files with sink extension
        sink_files = [
            file.with_suffix(sink_extension) for file in source_files]

    else:

        # Get sources and sinks as file lists
        source_files, sink_files = [[] for _ in sources], []
        for source_tuple, sink in zip(zip(*sources), sinks):
            source_paths = [Path(source) for source in source_tuple]
            source_is_dir = list(set([source.is_dir() for source in source_paths]))
            sink = Path(sink)

            if len(source_is_dir) > 1 and True in source_is_dir:
                raise ValueError('cannot handle directories when more than one input list is given')

            # Handle input directory (only if one source list)
            if True in source_is_dir:
                if not sink.is_dir():
                    raise RuntimeError(
                        f'For input tuple {source_tuple}, corresponding '
                        f'output {sink} is not a directory')
                for extension in source_extensions:
                    source_files[0] += list(source_tuple[0].rglob(f'*{extension}'))

                # Ensure one-to-one
                source_stems = [file.stem for file in source_files[0]]
                if not len(source_stems) == len(set(source_stems)):
                    raise ValueError(
                        'Two or more files have the same '
                        'stem with different extensions')

                # Get corresponding output files
                sink_files += [
                    sink / (file.stem + sink_extension)
                    for file in source_files]

            # Handle input file(s)
            else:
                if sink.is_dir():
                    raise RuntimeError(
                        f'For input file {source}, corresponding '
                        f'output {sink} is a directory')
                for i, source in enumerate(source_tuple):
                    source_files[i].append(source)
                sink_files.append(sink)

    if len(source_files) == 1:
        source_files = source_files[0]

    return source_files, sink_files



import contextlib
import time
from typing import Optional


###############################################################################
# Timer context
###############################################################################


@contextlib.contextmanager
def context(name: Optional[str]):
    """Wrapper to handle context changes of global timer

    Arguments
        name - Name of the timer to add time to
    """
    if not hasattr(context, 'timer'):
        context.timer = Timer()

    # TODO - support nested contexts
    context.timer.name = name
    with context.timer:
        yield
    context.timer.name = 'total'


def reset():
    """Clear timer state"""
    try:
        del context.timer
    except AttributeError:
        pass


def results() -> dict:
    """Get timing results

    Returns
        Timing results: {name: elapsed_time} for all names
    """
    if not hasattr(context, 'timer'):
        return {}
    return context.timer()


###############################################################################
# Utilities
###############################################################################


class Timer:
    """Context manager timer"""

    def __init__(self):
        self.reset()

    def __call__(self):
        """Retrieve timer results"""
        return {name: sum(times) for name, times in self.history.items()}

    def __enter__(self):
        """Start the timer"""
        self.start = time.time()

    def __exit__(self, *_):
        """Stop the timer"""
        elapsed = time.time() - self.start

        # Add to timer history
        if self.name not in self.history:
            self.history[self.name] = [elapsed]
        else:
            self.history[self.name].append(elapsed)

    def reset(self):
        """Reset the timer"""
        self.history = {}
        self.start = 0.
        self.name = 'total'



from typing import Iterable, Optional

import tqdm


###############################################################################
# TQDM iterator
###############################################################################


def iterator(
    iterable: Iterable,
    message: Optional[str],
    initial: int = 0,
    total: Optional[int] = None
) -> Iterable:
    """Create a tqdm iterator

    Arguments
        iterable
            Items to iterate over
        message
            Static message to display
        initial
            Position to display corresponding to index zero of iterable
        total
            Length of the iterable; defaults to len(iterable)
    """
    return tqdm.tqdm(
        iterable,
        desc=message,
        dynamic_ncols=True,
        initial=initial,
        total=len(iterable) if total is None else total)



import argparse
import os
from pathlib import Path
from typing import List, Optional, Union

import torchutil


###############################################################################
# Measure files and directories
###############################################################################


def main(
    globs: Union[str, List[str]],
    roots: Optional[
        List[
            Union[
                Union[str, bytes, os.PathLike],
                List[Union[str, bytes, os.PathLike]]
            ]
        ]
    ] = None,
    recursive: bool = False,
    unit='B'
) -> Union[int, float]:
    print(f'{torchutil.paths.measure(**locals())} {unit}')


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Measure data usage of files and directories')
    parser.add_argument(
        '--globs',
        required=True,
        nargs='+',
        help='Globs matching paths to measure')
    parser.add_argument(
        '--roots',
        type=Path,
        nargs='+',
        help='Directories to apply glob searches; '
              'current directory by default')
    parser.add_argument(
        '--recursive',
        action='store_true',
        help='Apply globs to all subdirectories of root directories')
    parser.add_argument(
        '--unit',
        choices=['B', 'KB', 'MB', 'GB', 'TB'],
        default='B',
        help='Unit of memory utilization (bytes to terabytes); default bytes')
    return parser.parse_args()


if __name__ == '__main__':
    main(**vars(parse_args()))



import contextlib
import os
from typing import Union


###############################################################################
# Path management utilities
###############################################################################


@contextlib.contextmanager
def chdir(directory: Union[str, bytes, os.PathLike]) -> None:
    """Context manager for changing the current working directory

    Arguments
        directory
            The desired working directory
    """
    curr_dir = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(curr_dir)



import os
from typing import Dict, Optional, Union

import accelerate
import torch
from torch.utils.tensorboard import SummaryWriter


###############################################################################
# Constants
###############################################################################


# Device managenent state
STATE = accelerate.state.PartialState()


###############################################################################
# Tensorboard logging
###############################################################################


@STATE.on_main_process
def update(
    directory: Union[str, bytes, os.PathLike],
    step: int,
    audio: Optional[Dict[str, torch.Tensor]] = None,
    sample_rate: Optional[int] = None,
    figures: Optional[Dict] = None,
    images: Optional[Dict[str, torch.Tensor]] = None,
    scalars: Optional[Dict[str, Union[float, int, torch.Tensor]]] = None):
    """Update Tensorboard

    Arguments
        directory - Directory to write Tensorboard files
        step - Training step
        audio - Optional dictionary of 2D audio tensors to monitor
        sample_rate - Audio sample rate; required if audio is not None
        figures - Optional dictionary of Matplotlib figures to monitor
        images - Optional dictionary of 3D image tensors to monitor
        scalars - Optional dictionary of scalars to monitor
    """
    if audio is not None:
        write_audio(directory, step, audio, sample_rate)
    if figures is not None:
        write_figures(directory, step, figures)
    if images is not None:
        write_images(directory, step, images)
    if scalars is not None:
        write_scalars(directory, step, scalars)


###############################################################################
# Writer
###############################################################################


@STATE.on_main_process
def writer(directory):
    """Get the writer object"""
    if not hasattr(writer, 'writer') or writer.directory != directory:
        writer.writer = SummaryWriter(log_dir=directory)
        writer.directory = directory
    return writer.writer


###############################################################################
# Utilities
###############################################################################


@STATE.on_main_process
def write_audio(directory, step, audio, sample_rate):
    """Write audio to Tensorboard"""
    for name, waveform in audio.items():
        writer(directory).add_audio(name, waveform, step, sample_rate)


@STATE.on_main_process
def write_figures(directory, step, figures):
    """Write figures to Tensorboard"""
    for name, figure in figures.items():
        writer(directory).add_figure(name, figure, step)


@STATE.on_main_process
def write_images(directory, step, images):
    """Write images to Tensorboard"""
    for name, image in images.items():
        writer(directory).add_image(name, image, step, dataformats='HCW')


@STATE.on_main_process
def write_scalars(directory, step, scalars):
    """Write scalars to Tensorboard"""
    for name, scalar in scalars.items():
        writer(directory).add_scalar(name, scalar, step)



import argparse
from pathlib import Path

import torchutil


###############################################################################
# Remove files and directories
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Remove files and directories')
    parser.add_argument(
        '--globs',
        required=True,
        nargs='+',
        help='Globs matching paths to delete')
    parser.add_argument(
        '--roots',
        type=Path,
        nargs='+',
        help='Directories to apply glob searches; '
              'current directory by default')
    parser.add_argument(
        '--recursive',
        action='store_true',
        help='Apply globs to all subdirectories of root directories')
    parser.add_argument(
        '--force',
        action='store_true',
        help='Skip user confirmation of deletion')
    return parser.parse_args()


if __name__ == '__main__':
    torchutil.paths.purge(**vars(parse_args()))



import os
import shutil
from pathlib import Path
from typing import List, Optional, Union

import torchutil


###############################################################################
# Purge
###############################################################################


def purge(
    globs: Union[str, List[str]],
    roots: Optional[
        List[
            Union[
                Union[str, bytes, os.PathLike],
                List[Union[str, bytes, os.PathLike]]
            ]
        ]
    ] = None,
    recursive: bool = False,
    force: bool = False
) -> None:
    """Remove all files and directories within directory matching glob

    Arguments
        globs
            Globs matching paths to delete
        roots
            Directories to apply glob searches; current directory by default
        recursive
            Apply globs to all subdirectories of root directories
        force
            Skip user confirmation of deletion
    """
    # Argument handling
    if isinstance(globs, str):
        globs = [globs]
    if not isinstance(roots, list):
        roots = [roots]
    roots = [Path() if root is None else Path(root) for root in roots]

    # Get paths to delete
    paths = []
    for root in roots:
        root = Path(root)
        for glob in globs:
            paths.extend(root.rglob(glob) if recursive else root.glob(glob))

    # Maybe prompt user
    if not force:
        files = len([path for path in paths if path.is_file()])
        dirs = len(paths) - files
        if input(
            f'Found {files} files and {dirs} directories to delete within '
            f'{len(roots)} root directories. Continue? [y/N]:'
        ).lower() != 'y':
            raise RuntimeError('Purge aborted by user')

    # Delete paths
    for path in torchutil.iterator(paths, 'Deleting paths'):
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()



import os
import requests
import tarfile
import tempfile
import zipfile
from typing import Union


###############################################################################
# Data download utilities
###############################################################################


def file(url: 'str', path: Union[str, bytes, os.PathLike]):
    """Download file from url

    Arguments
        url - The URL to download
        path - The location to save results
    """
    with requests.get(url, stream=True) as rstream:
        rstream.raise_for_status()
        with open(path, 'wb') as fstream:
            for chunk in rstream.iter_content(chunk_size=128):
                fstream.write(chunk)


def tarbz2(url: 'str', path: Union[str, bytes, os.PathLike]):
    """Download and extract tar bz2 file to location

    Arguments
        url - The URL to download
        path - The location to save results
    """
    with requests.get(url, stream=True) as rstream:
        rstream.raise_for_status()
        with tarfile.open(fileobj=rstream.raw, mode='r|bz2') as tstream:
            tstream.extractall(path)


def targz(url: 'str', path: Union[str, bytes, os.PathLike]):
    """Download and extract tar gz file to location

    Arguments
        url - The URL to download
        path - The location to save results
    """
    with requests.get(url, stream=True) as rstream:
        rstream.raise_for_status()
        with tarfile.open(fileobj=rstream.raw, mode='r|gz') as tstream:
            tstream.extractall(path)


def zip(url: 'str', path: Union[str, bytes, os.PathLike]):
    """Download and extract zip file to location

    Arguments
        url - The URL to download
        path - The location to save results
    """
    with tempfile.NamedTemporaryFile() as tmp:
        file(url, tmp.name)
        with zipfile.ZipFile(tmp, 'r') as zfile:
            zfile.extractall(path)



import math
from abc import ABC, abstractmethod
from typing import Tuple

import torch


###############################################################################
# Base batch-updating metric
###############################################################################


class Metric(ABC):
    """Base batch-updating metric"""

    @abstractmethod
    def __call__(self) -> float:
        """Retrieve the current metric values

        Returns
            The current metric value
        """
        pass

    @abstractmethod
    def update(self, predicted: torch.Tensor, target: torch.Tensor) -> None:
        """Update the metric

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        pass

    @abstractmethod
    def reset(self) -> None:
        """Reset the metric"""
        pass


###############################################################################
# Derived batch-updating metrics
###############################################################################


class Accuracy(Metric):
    """Batch-updating accuracy metric"""

    def __init__(self) -> None:
        self.reset()

    def __call__(self)-> float:
        """Retrieve the current accuracy value


            The current accuracy value
        """
        return (self.true_positives / self.count).item()

    def update(self, predicted: torch.Tensor, target: torch.Tensor) -> None:
        """Update accuracy

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        self.count += predicted.numel()
        self.true_positives += (predicted == target).sum()

    def reset(self) -> None:
        """Reset accuracy"""
        self.count = 0
        self.true_positives = 0


class Average(Metric):
    """Batch-updating average metric"""

    def __init__(self) -> None:
        self.reset()

    def __call__(self)-> float:
        """Retrieve the current average value


            The current average value
        """
        return (self.total / self.count).item()

    def update(self, values: torch.Tensor, count: int) -> None:
        """Update running average

        Arguments
            values
                The values to average
            count
                The number of values
        """
        self.count += count
        self.total += values.sum()

    def reset(self) -> None:
        """Reset running average"""
        self.count = 0
        self.total = 0.


class F1(Metric):
    """Batch-updating F1 score"""

    def __init__(self) -> None:
        self.precision = Precision()
        self.recall = Recall()

    def __call__(self) -> float:
        """Retrieve the current F1 value


            The current F1 value
        """
        precision, recall = self.precision(), self.recall()
        try:
            return 2 * precision * recall / (precision + recall)
        except ZeroDivisionError:
            return 0

    def update(self, predicted: torch.Tensor, target: torch.Tensor) -> None:
        """Update F1

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        self.precision.update(predicted, target)
        self.recall.update(predicted, target)

    def reset(self) -> None:
        """Reset F1"""
        self.precision.reset()
        self.recall.reset()


class L1(Metric):
    """Batch updating L1 score"""

    def __init__(self) -> None:
        self.reset()

    def __call__(self) -> float:
        """Retrieve the current L1 value


            The current L1 value
        """
        return (self.total / self.count).item()

    def update(self, predicted, target) -> None:
        """Update L1

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        self.count += predicted.numel()
        self.total += torch.abs(predicted - target).sum()

    def reset(self) -> None:
        """Reset L1"""
        self.count = 0
        self.total = 0.


class MeanStd(Metric):
    """Batch updating mean and standard deviation"""

    def __init__(self) -> None:
        self.reset()

    def __call__(self) -> Tuple[float, float]:
        """Retrieve the current mean and standard deviation

        Returns
            The current mean and standard deviation
        """
        std = math.sqrt(self.m2 / (self.count - 1))
        return self.mean, std

    def update(self, values: torch.Tensor) -> None:
        """Update mean and standard deviation

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        for value in values:
            self.count += 1
            delta = value - self.mean
            self.mean += delta / self.count
            delta2 = value - self.mean
            self.m2 += delta * delta2

    def reset(self) -> None:
        """Reset mean and standard deviation"""
        self.m2 = 0.
        self.mean = 0.
        self.count = 0


class PearsonCorrelation(Metric):
    """Batch-updating Pearson correlation"""

    def __init__(
        self,
        predicted_mean: float,
        predicted_std: float,
        target_mean: float,
        target_std: float
    ) -> None:
        """
        Arguments
            predicted_mean - Mean of predicted values
            predicted_std - Standard deviation of predicted values
            target_mean - Mean of target values
            target_std - Standard deviation of target values
        """
        self.reset()
        self.predicted_mean = predicted_mean
        self.predicted_std = predicted_std
        self.target_mean = target_mean
        self.target_std = target_std

    def __call__(self) -> float:
        """Retrieve the current correlation value

        Returns
            The current correlation value
        """
        return (
            1. / (self.predicted_std * self.target_std + 1e-6) *
            (self.total / self.count).item())

    def update(self, predicted, target) -> None:
        """Update Pearson correlation

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        self.total += sum(
            (predicted - self.predicted_mean) * (target - self.target_mean))
        self.count += predicted.numel()

    def reset(self) -> None:
        """Reset Pearson correlation"""
        self.count = 0
        self.total = 0.


class Precision(Metric):
    """Batch-updating precision metric"""

    def __init__(self) -> None:
        self.reset()

    def __call__(self) -> float:
        """Retrieve the current precision value

        Returns
            The current precision value
        """
        denominator = self.true_positives + self.false_positives
        return (self.true_positives / denominator).item()

    def update(self, predicted: torch.Tensor, target: torch.Tensor) -> None:
        """Update precision

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        self.true_positives += (predicted & target).sum()
        self.false_positives += (predicted & ~target).sum()

    def reset(self) -> None:
        """Reset precision"""
        self.true_positives = 0
        self.false_positives = 0


class Recall(Metric):
    """Batch-updating recall metric"""

    def __init__(self) -> None:
        self.reset()

    def __call__(self) -> float:
        """Retrieve the current recall value


            The current recall value
        """
        denominator = self.true_positives + self.false_negatives
        return (self.true_positives / denominator).item()

    def update(self, predicted: torch.Tensor, target: torch.Tensor) -> None:
        """Update recall

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        self.true_positives += (predicted & target).sum()
        self.false_negatives += (~predicted & target).sum()

    def reset(self) -> None:
        """Reset recall"""
        self.true_positives = 0
        self.false_negatives = 0


class RMSE(Metric):
    """Batch-updating RMSE metric"""

    def __init__(self) -> None:
        self.reset()

    def __call__(self) -> float:
        """Retrieve the current rmse value


            The current rmse value
        """
        return math.sqrt((self.total / self.count).item())

    def update(self, predicted: torch.Tensor, target: torch.Tensor) -> None:
        """Update RMSE

        Arguments
            predicted
                The model prediction
            target
                The corresponding ground truth
        """
        self.count += predicted.numel()
        self.total += ((predicted - target) ** 2).sum()

    def reset(self) -> None:
        """Reset RMSE"""
        self.count = 0
        self.total = 0.



from setuptools import find_packages, setup


with open('README.md') as file:
    long_description = file.read()


setup(
    name='torchutil',
    description='PyTorch utilities for developing deep learning frameworks',
    version='0.0.8',
    author='Max Morrison',
    author_email='maxrmorrison@gmail.com',
    url='https://github.com/maxrmorrison/torchutil',
    install_requires=['accelerate', 'apprise', 'tensorboard', 'torch'],
    packages=find_packages(),
    package_data={'torchutil': ['assets/*', 'assets/*/*']},
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['pytorch', 'utility', 'training'],
    classifiers=['License :: OSI Approved :: MIT License'],
    license='MIT')



import os
from typing import Callable, Dict, List, Optional, Tuple, Union

import torch


###############################################################################
# Checkpointing
###############################################################################


def load(
    file: Union[str, bytes, os.PathLike],
    model: torch.nn.Module,
    optimizer: Optional[torch.optim.Optimizer] = None,
    map_location: str = 'cpu'
) -> Tuple[
    torch.nn.Module,
    Union[None, torch.optim.Optimizer],
    Dict
]:
    """Load model checkpoint

    Arguments
        file - The checkpoint file
        model - The PyTorch model
        optimizer - Optional PyTorch optimizer for training
        map_location - The device to load the checkpoint on

    Returns
        model - The model with restored weights
        optimizer - Optional optimizer with restored parameters
        state - Additional values that the user defined during save
    """
    checkpoint = torch.load(file, map_location=map_location)

    # Restore model
    model.load_state_dict(checkpoint['model'])
    del checkpoint['model']

    # Restore optimizer
    if optimizer is not None:
        optimizer.load_state_dict(checkpoint['optimizer'])
    del checkpoint['optimizer']

    return model, optimizer, checkpoint


def save(
    file: Union[str, bytes, os.PathLike],
    model: torch.nn.Module,
    optimizer: torch.optim.Optimizer,
    accelerator=None,
    **kwargs
) -> None:
    """Save training checkpoint to disk

    Arguments
        file - The checkpoint file
        model - The PyTorch model
        optimizer - The PyTorch optimizer
        accelerator - HuggingFace Accelerator for device management
        kwargs - Additional values to save
    """
    if accelerator is None:
        save_fn = torch.save
    else:
        save_fn = accelerator.save
    checkpoint = {
        'model': model.state_dict(),
        'optimizer': optimizer.state_dict()}
    save_fn(checkpoint | kwargs, file)


###############################################################################
# Search predicates
###############################################################################


def highest_score(
    files: List[Union[str, bytes, os.PathLike]],
    scores: List[float]
) -> Tuple[Union[str, bytes, os.PathLike], float]:
    """Default for determining best checkpoint

    Arguments
        files - The checkpoint files

    Returns
        best_file - The filename of the checkpoint with the best score
        best_score - The corresponding score
    """
    index = torch.argsort(torch.tensor(scores))[-1]
    return files[index], scores[index]


def largest_number_filename(
    files: List[Union[str, bytes, os.PathLike]]
) -> Union[str, bytes, os.PathLike]:
    """Default for determining latest path; assumes filename is steps

    Arguments
        files - The checkpoint files

    Returns
        The filename containing the largest number
    """
    files.sort(key=lambda path: int(''.join(filter(str.isdigit, path.stem))))
    return files[-1]


###############################################################################
# Search checkpoints
###############################################################################


def best_path(
    directory: Union[str, bytes, os.PathLike],
    glob: str = '*.pt',
    best_fn: Callable = highest_score
) -> Tuple[Union[str, bytes, os.PathLike], float]:
    """Retrieve the path to the best checkpoint

    Arguments
        directory - The directory to search for checkpoint files
        glob - The glob matching the checkpoints
        best_fn - Takes a list of checkpoint paths and returns the latest
                  Default assumes checkpoint names are training step count.

    Returns
        best_file - The filename of the checkpoint with the best score
        best_score - The corresponding score
    """
    # Retrieve checkpoint filenames
    files = list(directory.glob(glob))

    # If no matching checkpoint files, no training has occurred
    if not files:
        return

    # Get all scores
    scores = [torch.load(file, map_location='cpu')['score'] for file in files]
    return best_fn(files, scores)


def latest_path(
    directory: Union[str, bytes, os.PathLike],
    glob: str = '*.pt',
    latest_fn: Callable = largest_number_filename,
) -> Union[str, bytes, os.PathLike]:
    """Retrieve the path to the most recent checkpoint in a directory

    Arguments
        directory - The directory to search for checkpoint files
        glob - The glob matching the checkpoints
        latest_fn - Takes a list of checkpoint paths and returns the latest
                    Default assumes checkpoint names are training step count.
    Returns
        The latest checkpoint in directory according to latest_fn
    """
    # Retrieve checkpoint filenames
    files = list(directory.glob(glob))

    # If no matching checkpoint files, no training has occurred
    if not files:
        return

    # Retrieve latest checkpoint
    return latest_fn(files)
